unit vfs;

interface

uses classes, sysutils;

type

  TBlock = record
    previous: int64;
    next: int64;
    data: pointer;
  end;
  PBlock = ^TBlock;

  TBuffer = array of byte; //used for copying pointer/buffers

  TFileRecord = packed record
    filename: string[255];
    filesize: int64;
    startblock: int64;
  end;
  PFileRecord = ^TFileRecord;

(*

  TvfsFileStream

  ala TFileStream maar dan werkend met bestand in vfs

  gebruik read write block evt met offset binnen block
  -- kijk naar read write data routine.

  lees lijst met gebruikte blocks in

  dus als je iets wilt inlezen op positie 1024
  dan kun je met blocksize uitrekenen in welk block dit is en op welke offset
  dus block x openen en naar juiste positie binnen block gaan.

*)

  TBlockList= array of integer;

  TvfsFileStream = class(TStream)
  private
    FSize: Integer;
    FPosition: Integer;
    FFileRecord: TFileRecord; //name / begin block / stream (file) size
    FBlockList: TBlockList; //blocks used by stream in order
    FDataFileP: TObject;//??
  protected
  public
    constructor Create(aName: string; aVFSFileSystem: TObject);
    function Read(var Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function GetSize: Int64; override;
    procedure SetSize(NewSize: LongInt); override;
  end;

  TVirtualFileSystem = class
  private
    FDataFile: TFileStream;
    FDataFileName: string;
    FBlockSize: integer;
    procedure WriteHeader();
    function ReadBlock(ablockid: int64; pBuf: pointer; var count: integer): int64;
    procedure WriteBlock(); overload;
    procedure WriteBlock(ablockid, anextblockid, aprevblockid: int64; pBuf: pointer; aSize: integer); overload;
    function NextFreeBlock(): int64;
    function SecondNextFreeBlock(): int64;
  protected
  public
    constructor Create();
    destructor Destroy(); override;
    procedure New();
    procedure Open();
    procedure Close();

    //--------------------------------------------------------------------------

    function FindFile(aname: string): TFileRecord;
    procedure AddFile(aname: string; astartblockid: int64; asize: int64);

    //--------------------------------------------------------------------------

    function ReadData(aname: string; pBuf: pointer): integer;
    procedure WriteData(aname: string; pBuf: pointer; ASize: integer);

    function  ReadStream(Section, Ident: string; Stream: TStream): Integer;
    function  WriteStream(Section, Ident: string; Stream: TStream): Integer;

    //--------------------------------------------------------------------------

    function GetBlockList(afile: TFileRecord): TBlockList;
    function ReadVirtualFileStreamBuffer(var buffer; Position, Count: Integer; BlockList: TBlockList ): LongInt;
    function WriteVirtualFileStreamBuffer(const buffer; Position, Count: Integer; var BlockList: TBlockList ): LongInt;

    //--------------------------------------------------------------------------

    procedure WriteDummyRecord();
    procedure WriteDummyLarge();
    function ReadTest: string;
    procedure ReadDummyLarge();
    function ReadDummyRecord(): string;

    //--------------------------------------------------------------------------

    property FileName: string read FDataFileName write FDataFileName;
    property BlockSize: integer read FBlockSize;
  end;

implementation

//TvfsFileStream ---------------------------------------------------------------

constructor TvfsFileStream.Create(aname: string; aVFSFileSystem: TObject);
begin
  inherited Create();

  if aVFSFileSystem is TVirtualFileSystem then
  begin
    self.FDataFileP := aVFSFileSystem;

    //Bestaat aname reeds?  //evt bestaande bestand verwijderen...
    FFileRecord := TVirtualFileSystem(self.FDataFileP).FindFile(aname);
if FFileRecord.startblock > 0 then
begin
  //overschrijven huidige bestand (dus verwijderen)
  FSize:=FFileRecord.filesize;
end
else
begin
  TVirtualFileSystem(self.FDataFileP).AddFile(aname,TVirtualFileSystem(self.FDataFileP).NextFreeBlock() , 0);
  FSize:=0;
end;




    FPosition:=0;
    FBlockList:=TVirtualFileSystem(self.FDataFileP).GetBlockList(FFileRecord);
    FPosition:=0;
  end
  else
    Raise Exception.Create('aVFSFileSystem must be a TVirtualFileSystem');

end;

function TvfsFileStream.GetSize(): Int64;
begin
  result :=FSize;
end;

procedure TvfsFileStream.SetSize(NewSize: LongInt);
begin
  // Set the size property of the wrapped stream
  FSize:= NewSize;
  FFileRecord.FileSize := FSize;
  //currently overwrites first and only dir entry
  TVirtualFileSystem(self.FDataFileP).WriteBlock(1,0,0,@FFileRecord, SizeOf(TFileRecord));
end;


function TvfsFileStream.Read(var Buffer; Count: Integer): Longint;
begin

//naar functie in TVirtualFileSystem

//TVirtualFileSystem.ReadStream(var buffer; Position, Count: Integer): LongInt;
result := TVirtualFileSystem(FDataFileP).ReadVirtualFileStreamBuffer(Buffer, FPosition, Count, FBlockList);

end;

function TvfsFileStream.Seek(Offset: Integer; Origin: Word): Longint;
var
  len : integer;
begin
  len := FFileRecord.filesize;
  case Origin of
    soFromBeginning: FPosition := Offset;
    soFromCurrent: FPosition := FPosition + Offset;
    soFromEnd: FPosition := len - Offset;
  end;
  if (FPosition > len) then
  begin
    FPosition := len;
  end else
  if (FPosition < 0) then
  begin
    FPosition := 0;
  end;
  Result := FPosition;
end;

function TvfsFileStream.Write(const Buffer; Count: Integer): Longint;
var
  MyBuffer: Pointer;
begin
//  result := 0;

  result := TVirtualFileSystem(FDataFileP).WriteVirtualFileStreamBuffer(Buffer, FPosition, Count, FBlockList);

  //TODO: detect if size of stream has changed?
  // Set the size property of the wrapped stream


end;

//TvfsFileSystem ---------------------------------------------------------------

constructor TVirtualFileSystem.Create();
begin
  FDataFile := nil;
  FBlockSize := 512;
end;

destructor TVirtualFileSystem.Destroy();
begin
  if FDataFile <> nil then self.Close();
  inherited Destroy();
end;

procedure TVirtualFileSystem.New();
begin
  FDataFile:= TFileStream.Create(FDataFileName, fmCreate);
  //self.WriteBlock();
  self.WriteHeader();
  self.WriteDummyRecord();
  //self.WriteDummyLarge();
  self.Close();
  self.Open();
end;

procedure TVirtualFileSystem.Open();
begin
  FDataFile := TFileStream.Create(FDataFileName, fmOpenReadWrite);
  FDataFile.Position := 0;
end;

procedure TVirtualFileSystem.Close();
begin
  FDataFile.Free;
end;

procedure TVirtualFileSystem.WriteBlock(); //depracted do not use
var
  block : TBlock;
  filerecord: PFileRecord;
  i: integer;
begin
  system.New(filerecord);
  fillchar(filerecord.filename, SizeOf(filerecord.filename),0);

  filerecord.filename := 'root folder';
  filerecord.filesize := 0;
  filerecord.startblock := 0;

  block.previous := 0;
  block.next := 0;
  GetMem(block.data, FBlockSize);
  FillChar(block.data^, FBlockSize, 0);

  for i:=0 to sizeof(tfilerecord)-1 do
  begin
    TBuffer(block.data)[i]:=TBuffer(filerecord)[i];
  end;

  //Move(filerecord, block.data, sizeof(tfilerecord) );


  FDataFile.Position :=0;
  FDataFile.WriteBuffer(block.previous,sizeof(block.previous));
  FDataFile.WriteBuffer(block.next,sizeof(block.next));
  FDataFile.WriteBuffer(block.data^,FBlockSize);

  freemem(block.data);
  system.Dispose(filerecord);
end;

function TVirtualFileSystem.ReadBlock(ablockid: int64; pBuf: pointer; var count: integer): int64;
var
  block: TBlock;
  i: integer;
begin
  GetMem(block.data, FBlockSize - (sizeof(int64)*2) );

  FDataFile.Position :=(ablockid * FBlockSize);// + (sizeof(int64)*2);
  FDataFile.ReadBuffer(block.previous, sizeof(block.previous));
  FDataFile.ReadBuffer(block.next, sizeof(block.next));
  FDataFile.ReadBuffer(block.data^, FBlockSize - (sizeof(int64)*2) );

  //how to detect if block is not filled completely? Need to know filesize in ReadData function
  for i:=0 to FBlockSize-1- (sizeof(int64)*2) do
  begin
    TBuffer(pBuf)[i]:=TBuffer(block.data)[i];
  end;
  FreeMem(block.data);

  result := block.next;
end;

function TVirtualFileSystem.ReadTest: string;
var
  block : TBlock;
  filerecord: PFileRecord;
  i: integer;
begin
  system.new(filerecord);
  GetMem(block.data, FBlockSize- (sizeof(int64)*2) );
  FillChar(block.data^, FBlockSize- (sizeof(int64)*2), 0);

  FDataFile.Position :=(0 * FBlockSize);// + (sizeof(int64)*2);
  FDataFile.ReadBuffer(block.previous, sizeof(block.previous));
  FDataFile.ReadBuffer(block.next, sizeof(block.next));
  FDataFile.ReadBuffer(block.data^, FBlockSize- (sizeof(int64)*2));


  for i:=0 to sizeof(tfilerecord)-1 do
  begin
    TBuffer(filerecord)[i]:=TBuffer(block.data)[i];
  end;

  //Move(block.data, filerecord, sizeof(filerecord));

  result := filerecord.filename + ' ' + IntToStr(self.NextFreeBlock);

  FreeMem(block.data);
  system.Dispose(filerecord);
end;

//block 0 voor de zekerheid gebruiken voor header etc....

//http://nl.wikipedia.org/wiki/FAT16
//next en prev meegeven in aanroep.
procedure TVirtualFileSystem.WriteBlock(ablockid, anextblockid, aprevblockid: int64; pBuf: Pointer; aSize: integer);
var
  Block: TBlock;
  i: integer;
begin
  block.data := nil;
  //hoe bepaal ik waar dit block wordt weggeschreven?  block id ?
  block.previous := aprevblockid; //how to determine | 0 if first block for file
  block.next := anextblockid; //how to determine ( getnextfreeblock function ? ) maar deze kan dus dit huidige block teruggeven omdat deze nog niet geschreven is?

  i:=FBlockSize-(sizeof(int64)*2);
  system.GetMem(block.data, i );
  FillChar(block.data^, FBlockSize-(sizeof(int64)*2) , 0);

  for i:=0 to aSize-1 do
  begin
    TBuffer(block.data)[i]:=TBuffer(pBuf)[i];
  end;

  FDataFile.Position := (ablockid * FBlockSize);// + (sizeof(int64)*2);

  FDataFile.WriteBuffer(block.previous,sizeof(block.previous));
  FDataFile.WriteBuffer(block.next,sizeof(block.next));

  FDataFile.WriteBuffer(block.data^,FBlockSize-(sizeof(int64)*2) );

  system.freemem(block.data);
end;

function TVirtualFileSystem.ReadData(aname: string; pBuf: pointer): integer;
var
  size, count: integer;
  prevblock, nextblock: int64;
  buffer: pointer;
  i: integer;
  FileRecord: TFileRecord;
begin

  FileRecord := findfile(aname);
  if FileRecord.startblock > 0 then
  begin

  //system.getmem(pbuf, FileRecord.filesize);
  system.getmem(buffer, FBlockSize - (sizeof(int64)*2) );
  //ophalen eerste te lezen block via aname
  nextblock := FileRecord.startblock; //voor test op block 1
  prevblock := 0;
  size := 0;
  while nextblock <> 0 do
  begin

    count := FBlockSize -(sizeof(int64)*2);
    nextblock := ReadBlock(nextblock, buffer, count );

    for i:=0 to FBlockSize -1-(sizeof(int64)*2) do
    begin
      if i+size < Filerecord.filesize then
        TBuffer(pBuf)[i+size]:=TBuffer(buffer)[i];
    end;

    size := size + FBlockSize - (sizeof(int64)*2);
  end;

  system.freemem(buffer);

  end;

  result := FileRecord.filesize;

end;

procedure TVirtualFileSystem.WriteData(aname: string; pBuf: pointer; aSize: integer);
var
  buffer: pointer;
  i: integer;
  tempsize: integer;
  prevblock, nextblock: int64;
  blockcount: integer;
begin
prevblock := 0;

//Bestaat aname reeds?  //evt bestaande bestand verwijderen...
if findfile(aname).startblock > 0 then
begin
  //overschrijven huidige bestand (dus verwijderen)
end
else
begin
  AddFile(aname,self.NextFreeBlock() , aSize);
end;

//controle of size van pBuf
//evt verdelen over meerdere blokken van FBlockSize grootte;

if aSize > (FBlockSize - (sizeof(int64)*2) )then
begin

  tempsize := aSize;
  blockcount := 0;

  while tempsize > (FBlockSize - (sizeof(int64)*2) ) do
  begin
    system.GetMem(buffer, FBlockSize - (sizeof(int64)*2) );
    fillchar(buffer^, FBlockSize - (sizeof(int64)*2) , 0); //hier evt overbodig...
    for i:=0 to FBlockSize-1-(sizeof(int64)*2) do
    begin
      TBuffer(buffer)[i] := TBuffer(pBuf)[i+blockcount];
    end;
    nextblock := self.NextFreeBlock();
    self.WriteBlock(nextblock,self.SecondNextFreeBlock(),prevblock,buffer, FBlockSize-(sizeof(int64)*2));
    prevblock := nextblock;
    tempsize := tempsize - FBlockSize + (sizeof(int64)*2);
    system.FreeMem(buffer);

    blockcount := blockcount + FBlockSize-(sizeof(int64)*2);//16;
  end;

  //fix to use correct last buffer
  system.GetMem(buffer, FBlockSize-(sizeof(int64)*2));
  fillchar(buffer^, FBlockSize-(sizeof(int64)*2), 0); //hier evt overbodig...
  for i:=0 to tempsize-1 do
  begin
    TBuffer(buffer)[i] := TBuffer(pBuf)[i+blockcount];
  end;
  self.WriteBlock(self.NextFreeBlock(),0,prevblock,buffer, FBlockSize-(sizeof(int64)*2));
  system.FreeMem(buffer);
end
else
begin
  //just write the complete buffer if it is smaller then the blocksize
  system.GetMem(buffer, FBlockSize-(sizeof(int64)*2));
  fillchar(buffer^, FBlockSize-(sizeof(int64)*2), 0); //hier evt overbodig...
  for i:=0 to aSize-1 do
  begin
    TBuffer(buffer)[i] := TBuffer(pBuf)[i];
  end;
  self.WriteBlock(self.NextFreeBlock(),0,prevblock,buffer, FBlockSize-(sizeof(int64)*2));
  system.FreeMem(buffer);
end;

end;

function TVirtualFileSystem.NextFreeBlock(): int64;
begin
  //bepaal het volgende vrije block (filesize / FBlockSize);
  //moet ook kijken naar evt verwijderde/vrije blokken
  //denk ook om overschrijven zelfde bestand ...
  result := (FDataFile.Size div FBlockSize);
end;

function TVirtualFileSystem.SecondNextFreeBlock(): int64;
begin
  //bepaal het volgende vrije block (filesize / FBlockSize);
  //moet ook kijken naar evt verwijderde/vrije blokken
  result := (FDataFile.Size div FBlockSize)+1;
end;

procedure TVirtualFileSystem.WriteDummyLarge();
var
  data: pointer;
  size: integer;
begin
  size:=(1024*2)-256;
  system.GetMem(data, size);
  fillchar(data^, size, 1);
  self.WriteData('lt', data, size);
  system.freemem(data);
end;

procedure TVirtualFileSystem.WriteDummyRecord();
var
  FileRecord : TFileRecord;
begin
  fillchar(filerecord, SizeOf(TFileRecord),0); //clean up memory
  FileRecord.filename := 'Root Folder';
  FileRecord.startblock := self.NextFreeBlock(); //of 1
  //self.WriteData('/',@FileRecord, SizeOf(TFileRecord));
  self.AddFile('Root Folder', 1, 0);
end;

function TVirtualFileSystem.ReadDummyRecord(): string;
var
  FileRecord : TFileRecord;
begin
  //system.New(FileRecord);
  self.ReadData('/', @FileRecord);
  result := FileRecord.filename;
end;

procedure TVirtualFileSystem.ReadDummyLarge();
var
  data: pointer;
  size: integer;
begin
  size:=(1024*2);//-256;
  system.GetMem(data, size);
//  fillchar(data^, size, 0);
  self.ReadData('lt', data);

  //

  system.freemem(data);
end;

function  TVirtualFileSystem.ReadStream(Section, Ident: string; Stream: TStream): Integer;
var
  buffer: pointer;
  size: int64;
    FileRecord: TFileRecord;

begin
 FileRecord := findfile(Ident);

  system.getmem(buffer, FileRecord.filesize);

  size:=self.ReadData(Ident, buffer);
  stream.Write(buffer^, size);

  system.freemem(buffer);

  result := FileRecord.filesize;
end;

function  TVirtualFileSystem.WriteStream(Section, Ident: string; Stream: TStream): Integer;
var
  buffer: pointer;
//  size: integer;

begin



  system.GetMem(buffer,stream.Size);

  stream.Read(buffer^,stream.Size);
  self.WriteData(Ident, buffer, stream.Size);

  system.FreeMem(buffer);

  result := stream.Size;
end;

procedure TVirtualFileSystem.WriteHeader();
var
  FileRecord : TFileRecord;
begin
  fillchar(filerecord, SizeOf(TFileRecord),0); //clean up memory
  FileRecord.filename := 'Header Data';
  FileRecord.startblock := 0;

  self.WriteBlock(0,0,0,@FileRecord, SizeOf(TFileRecord));
//  self.WriteData('/header/',@FileRecord, SizeOf(TFileRecord));
end;

function TVirtualFileSystem.FindFile(aname: string): TFileRecord;
var
  FileRecord: TFileRecord;
  buffer: pointer;
  i,size: integer;
begin
  FileRecord.filename := 'Not Found';
  FileRecord.filesize := 0;
  FileRecord.startblock := 0;
  result :=FileRecord;

  //read first dir entry
  system.GetMem(buffer, FBlockSize-(sizeof(int64)*2) );
  size:= FBlockSize-(sizeof(int64)*2);
  self.ReadBlock(1,buffer, size );
  for i:=0 to sizeof(tfilerecord)-1 do
  begin
    TBuffer(@filerecord)[i]:=TBuffer(buffer)[i];
  end;
  system.FreeMem(buffer);

  //compare
  if FileRecord.filename = aname then
  begin
    result := FileRecord;
  end;

end;

procedure TVirtualFileSystem.AddFile(aname: string; astartblockid: int64; aSize: int64);
var
  FileRecord: TFileRecord;
begin
  fillchar(filerecord, SizeOf(TFileRecord),0); //clean up memory
  FileRecord.filename := aname;
  FileRecord.startblock := astartblockid;
  FileRecord.filesize := aSize;

  //currently overwrites first and only dir entry
  self.WriteBlock(1,0,0,@FileRecord, SizeOf(TFileRecord));
end;

//------------------------------------------------------------------------------

function TVirtualFileSystem.GetBlockList(afile: TFileRecord): TBlockList;
var
  i: integer;
  totalblocks: integer;
  p, n: int64;
  SIZE: integer;
  FILESIZE: integer;
begin
  //FBlockList vullen met gebruikte blocks moet naar TVirtualFileSystem
  //TVirtualFileSystem.GetBlockList(): TBlockList;

  FILESIZE:=afile.filesize;
  SIZE:= FBlockSize - (2*SizeOf(int64) );

  //totalblocks berekenen klopt niet omdat filesize niet overeenkomt met blockgrootte...
  totalblocks := (FileSize div SIZE)+2; //waarom moeten er 2 bij?
  //SetLength(FBlockList, totalblocks+2);


   if afile.filename = 'Not Found' then
    result := nil
   else
   begin
    SetLength(result, 1);
    n:=afile.startblock;
    result[0] := n;
   end;

  i:=1;
  //for i:=0 to totalblocks-1 do
  if FILESIZE > SIZE then
  begin
    while n <> 0 do
    begin
      SetLength(result, i+1);
      FdataFile.Position := n * FBlockSize ;
      FDataFile.ReadBuffer(p, sizeof(p));
      FDataFile.ReadBuffer(n, sizeof(n));
      result[i] := n;
      i:=i+1;
    end;
  end;
end;

function TVirtualFileSystem.ReadVirtualFileStreamBuffer(var buffer; Position, Count: Integer; BlockList: TBlockList): LongInt;
var
   buf_pos: PChar;
   ibegin, iend: integer;
   iBeginBlock, iEndBlock: integer;
   SIZE: integer;
   i,t: integer;
   readbuffer: pointer;
   mySize: integer;
   b: int64;
begin

  buf_pos := pchar(@Buffer);
  result := count;

  ibegin := Position; //positie in bestandsstream
  iEnd   := Position+count; //einde in te lezen stream


  //Dec(Count, result);

  //buf_pos^ := 'X'; //yes the x gets written/read...

  //implement code from ReadData here ...  instead of 1st block
  SIZE:= FBlockSize - (2*SizeOf(int64) );

  //begin en eind block uitrekenen
  iBeginBlock := iBegin div FBlockSize;
  //iEndBlock klopt niet? nu wel met correctie?
  iEndBlock := iEnd div Size;//(TVirtualFileSystem(self.FDataFileP).FBlockSize);

  system.GetMem(readbuffer, SIZE); //

  mySize:=0;

  if iBeginBlock <> iEndBlock then
  begin

    for i:=iBeginBlock to iEndBlock do
    begin
      b:=BlockList[i];
      self.ReadBlock(b, readbuffer, SIZE);

      for t:=0 to SIZE-1 do
      begin
      if (i=0) and (t+mysize>=iBegin) then
        TBuffer(buf_pos)[t+mysize]:=TBuffer(readbuffer)[t];

      if (i>0) AND (i<iEndBlock) then
        TBuffer(buf_pos)[t+mysize]:=TBuffer(readbuffer)[t];

      if (i=iEndBlock) then
        if t+mysize+iBegin < iEnd then
          TBuffer(buf_pos)[t+mysize]:=TBuffer(readbuffer)[t];
      end;

      mysize := mysize + SIZE;
    end;

  end
  else
  begin
    //smaller then count so read only first block
    self.ReadBlock(BlockList[iBeginBlock], buf_pos, SIZE);
    //copy part from iBegin to iEnd from buffer;

  end;

  system.FreeMem(readbuffer);
end;

function TVirtualFileSystem.WriteVirtualFileStreamBuffer(const buffer; Position, Count: Integer; var BlockList: TBlockList ): LongInt;
var
  writebuffer: pointer;
  i,t: integer;
  tempsize: integer;
  prevblock, nextblock: int64;
  blockcount: integer;

     ibegin, iend: integer;
   iBeginBlock, iEndBlock: integer;
   buf_pos: PChar;
   Size, mysize: integer;
   b,nb: int64;
begin
prevblock := 0;

//controle of size van pBuf
//evt verdelen over meerdere blokken van FBlockSize grootte;
  buf_pos := pchar(@Buffer);
  result := count;

  ibegin := Position; //positie in bestandsstream
  iEnd   := Position+count; //einde in te lezen stream


  //Dec(Count, result);

  //buf_pos^ := 'X'; //yes the x gets written/read...

  //implement code from ReadData here ...  instead of 1st block
  SIZE:= FBlockSize - (2*SizeOf(int64) );

  //begin en eind block uitrekenen
  iBeginBlock := iBegin div FBlockSize;
  //iEndBlock klopt niet? nu wel met correctie?
  iEndBlock := iEnd div Size;//(TVirtualFileSystem(self.FDataFileP).FBlockSize);

  system.GetMem(writebuffer, SIZE); //

   mySize:=0;

  if iBeginBlock <> iEndBlock then
  begin

    for i:=iBeginBlock to iEndBlock do
    begin


//      if BlockList = nil then
//      begin
//        b:=self.NextFreeBlock();
//        nb:=self.SecondNextFreeBlock();
//        SetLength(BlockList, High(BlockList)+1);
//        BlockList[i] := b;
//        fillchar(writebuffer^, SIZE, 0); //clean the write buffer
//      end
//      else
      if i >High(BlockList) then
      begin
        b:=self.NextFreeBlock();
        nb:=self.SecondNextFreeBlock();
        if BlockList = nil then
          SetLength(BlockList, 1)
        else
          SetLength(BlockList, High(BlockList)+1);
        BlockList[i] := b;
        fillchar(writebuffer^, SIZE, 0); //clean the write buffer
      end
      else
      begin
        b:=BlockList[i];
        if (i+1) > High(BlockList) then
        begin
          nb:=self.NextFreeBlock;
        end
        else
        begin
          nb:=BlockList[i+1]; //detect i > iend or else;
        end;

        self.ReadBlock(b, writebuffer, SIZE); //read current content of buffer
      end;

      for t:=0 to SIZE-1 do
      begin
      if (i=0) and (t+mysize>=iBegin) then
      TBuffer(writebuffer)[t] := TBuffer(buf_pos)[t+mysize];

      if (i>0) AND (i<iEndBlock) then
        TBuffer(writebuffer)[t]:=TBuffer(buf_pos)[t+mysize];

      if (i=iEndBlock) then
        if t+mysize+iBegin < iEnd then
          TBuffer(writebuffer)[t]:=TBuffer(buf_pos)[t+mysize];
      end;

      self.WriteBlock(b,nb,0,writebuffer,SIZE);

      mysize := mysize + SIZE;
    end;

  end
  else
  begin
    if i > High(BlockList) then
      begin
        b:=self.NextFreeBlock();
      end
      else
      begin
        b:=BlockList[iBeginBlock];
      end;

    //smaller then count so read only first block //fix to write correct size of data?
    self.WriteBlock(b,0,0, buf_pos, SIZE);
    //copy part from iBegin to iEnd from buffer;

  end;

  system.FreeMem(writebuffer);


end;

end.
